/* march.c	Tesselate a dense array of data
 * This is an implementation of the "marching cubes" algorithm by Lorensen and
 * Cline at General Electric, presented in the SIGGRAPH '87 Conference
 * Proceedings.
 * This algorithm will take a dense array of data (type determined by Input
 * in march.h- hopefully floating point, as I'm not sure it'd work right if you
 * set it to integer values) and tesselate it at the specified functional value.
 * It currently consists of one step that scans the data and a second one that
 * does the actual tessellation.  The principle it operates on is that from zero to
 * four triangles can be used to describe the intersection of a surface with a
 * voxel, and so it deals with the surfaces presented at that level of abstraction.
 * This would lend itself well to parallel processing.
 * This specific routine handles the business of coordinating the marching cubes
 * process.  This file is intimately linked with array.c and tessel.c.
 *
 */

#include "march.h"

Big march(dx, dy, dz, value, array, edgelist, normlist, connlist, edges, sides, p_edges)
     unsigned long dx, dy, dz;	/* Data dimensions */
     Input value, ***array;	/* Array of data */
     Output *edgelist[3], *normlist[3]; /* Output storage */
     Big **connlist;		/* Output storage */
     int *edges, *sides;
     Big *p_edges;

{
    struct cube **cubes;
    Big *depths;
    Big num_triangles, pred_edges;
    Big edges_done, triangles_done, num_sides;
    int i, norm;
  

    if(normlist != NULL)
	norm = 1;
    else 
	norm = 0;


    if(!initialization_done)
	init_marchcubes();

#ifdef CHECKMALLOC
    (void) malloc_debug(2);	/* Maximum debugging level */
#endif

/* Set up arrays for storing the locations of relevant cubes.
 * It's better than just skipping through a large array.
 */
    if((cubes = (struct cube **) Calloc(dx-1, sizeof(struct cube *))) == NULL)
	perror("cubes");
    if((depths = (Big *) Calloc(dx-1, sizeof(Big))) == NULL)
	perror("depths");

    num_triangles = process_array(dx, dy, dz, value, array, cubes, 
				  depths, &pred_edges, &num_sides);

/* Internally, we avoid redundancy in the edge list by
 * checking for conflicts; 0 means "unused", so we index
 * from 1 and must make room for a zero element.
 */
    pred_edges++;
	
/* All the tessellation algorithms are similar:	 gouraud_tess()
 * and constant_tess() supply the same triangle information,
 * but the former function supplies normals as well.  fake_tess()
 * doesn't interpolate, giving you a really bizarre looking
 * picture; for very high resolution (100x100x100 or higher)
 * constant shaded surfaces look just like Gouraud shaded
 * surfaces.
 */
    if((*connlist = (Big *) Calloc(num_sides, sizeof(Big))) == NULL)
	perror("connlist");

    if(norm)		/* Need normals */
	gouraud_tess(array, value, dx, dy, dz, cubes, depths, edgelist, normlist, 
		     *connlist, pred_edges, &triangles_done, &edges_done);
    else
	constant_tess(array, value, dx, dy, dz, cubes, depths, edgelist, 
		      *connlist, pred_edges, &triangles_done, &edges_done);

/* Free unused information immediately.
 */
    for(i = 0; i < (dx-1); i++)
	if(depths[i])
	    Cfree(cubes[i], (unsigned)depths[i], sizeof(struct cube));
    Cfree(cubes, dx-1, sizeof(struct cube *));
    Cfree(depths, dx-1, sizeof(Big));

    *edges = edges_done;
    *sides = num_sides;
    *p_edges = pred_edges;
#ifdef DEBUG
    printf("Marching cubes complete:  %d triangles.\n", num_triangles);
#endif

    return(num_triangles);
}



